% =========================================================================
% Title       : Complex Reduction of Amplitudes for Parseval Frames (CRAMP)
% File        : CRAMP.m
% -------------------------------------------------------------------------
%
% Description:
%
%   Convex Reduction of Amplitudes (CRAMP) for Parseval frames (CRAMP). The
%   method solves 
%
%      argmin norm(x,'infinity')  subject to  Dx=b
%
%   where D is a MxN Parseval frame. The routine is described in 
%
%      C. Studer, T. Goldstein, W. Yin, and R. G. Baraniuk, "Democratic 
%      Representations," arXiv preprint: 1401.3420, 2014
%
% Usage:
%
%   [ solution, outs ] = CRAMP(D,Dt,b,maxIterations)
%
% Inputs:
%
%     D and Dt      :  Function handles to an MxN matrix D and its adjoint
%                      Dt, with M<N.  NOTE:  D MUST BE A TIGHT FRAME
%     maxIterations : limits number of iterations 
%
% Outputs:
%
%     solution : the minimizers of the energy
%     outs     : a struct containing various outputs and convergence data
%       outs.r : residuals showing convergence
%       outs.p : peak-to-average power ratio of each iterate
%
% ------------------------------------------------------------------------- 
%
% Revisions:
%
%   Date       Version  Author     Description
%   Jun-08-15  1.2      studer     cleanup for code release
%   Nov-13-13  1.0      goldstein  created
%
% -------------------------------------------------------------------------
% (C) 2015 Tom Goldstein (e-mail: tomg@cs.umd.edu)
% =========================================================================

function [ solution, outs ] = CRAMP(D,Dt,b,maxIterations)

%  Parameters set by the user
%  maxIterations = 100;  %  maximum Douglas Rachford iterations
tolerance = 1e-8;        %  stopping tolerance

%%  Check preconditions
%   verify that b is a column vector
assert(size(b,2)==1 & ndims(b)<3,'Input b must be a column vector');

% Verify that D*Dt is an identity (true for tight frames)
Dtb = Dt(b);
DDtb = D(Dtb);
assert(norm(b - DDtb,'fro')<1e-9,'D is not a Parseval frame');

%% Allocate memory
n = length(Dt(b)); %  problem size
m = length(b);
xp1 = Dtb;  %  initial iterates
zp1 = Dtb;
r = zeros(maxIterations,1); % residuals at each step. Used to measure convergence
p = zeros(maxIterations,1); % peak-to-average power ratio of iterates

%% steps size selection
peak = max(abs(Dtb));  %  Largest signal value
thresh = norm(b)/sqrt(n-.75*m); %  Estimated threshold for L_infinity
tau = 10*(peak)*(n-m);  %  Estimated amount of shrinkage needed to crush n-m values down to the peak.  Multiply by 10 cuz "why not?!"

best_papr = papr(xp1);
best_solution = xp1;

%% Perform Douglas Rachford Iteration
for iter=1:maxIterations
    x = xp1;
    z = zp1;
    xhat = prox_inf(z,tau);
    %xhat = chop(z,1.2*thresh); % a shortcut method with no sorting
    zhat = 2*xhat-z;
    xp1 = zhat - Dt(D(zhat)-b);
    zp1 = z+1.95*(xp1-xhat); %1.9
    
    %  Compute and store normalized residuals (convergence measure)
    residual1 = (xhat - z)/tau;
    residual2 = (xp1-zhat)/tau;
    r(iter) =  norm(residual1+residual2)/max(norm(residual1),norm(residual2));
    %  record PAPR
    p(iter) = papr(xp1);
    
    %  The method is not monotone in PAPR, so make sure we return the best
    %  one
    if p(iter)<best_papr
        best_papr = p(iter);
        best_solution = xp1;
    end
    
    %  If convergence is reached, record results and terminate
    if r(iter)<tolerance || iter == maxIterations
        outs = [];
        outs.r = r(1:iter);
        outs.p = p(1:iter);
        outs.sol = best_solution;
        solution = best_solution;
        outs.tau = tau;
        outs.m = m;
        outs.n = n;
        return;
    end
     
end

end


% Peak to average power ratio
function p = papr( w )
 N = length(w);
 p = N*norm(w,'inf')^2/norm(w,'fro')^2;
end